import { s } from './h.js'

const COLORS = ['#e22', '#e73', '#fc3', '#ad4', '#4d9', '#3be', '#45d', '#c3e']
const PADDING = 15

/**
 * A helper for creating a chart based on a table (conf. <https://jackens.github.io/nnn/chartable/>).
 *
 * Options:
 * - `bottom`: bottom padding (for X axis labels)
 * - `gapX`: X axis spacing
 * - `gapY`: Y axis spacing
 * - `headerColumn`: flag indicating that `table` has a header column (with X axis labels)
 * - `left`: left padding
 * - `maxY`: number of Y axis lines
 * - `reverse`: flag to reverse all data series
 * - `right`: right padding (for data series labels)
 * - `rotate`: flag to rotate X axis labels
 * - `table`: `HTMLTableElement` to extract data, data series labels and X axis labels
 * - `top`: top padding
 *
 * @param {{
 *   bottom?: number;
 *   gapX?: number;
 *   gapY?: number;
 *   headerColumn?: boolean;
 *   left?: number;
 *   maxY?: number;
 *   reverse?: boolean;
 *   right?: number;
 *   rotate?: boolean;
 *   table: HTMLTableElement;
 *   top?: number;
 * }} options
 */
export const chartable = ({
  bottom = 50,
  gapX = 70,
  gapY = 30,
  headerColumn = false,
  left = 300,
  maxY = 10,
  reverse = false,
  right = 100,
  rotate = false,
  table,
  top = PADDING
}) => {
  const /** @type {number[][]} */ zxValues = []
  const /** @type {string[]}   */ xLabels = []
  const /** @type {string[]}   */ zLabels = []

  table.querySelectorAll('tr').forEach((row, r) =>
    // @ts-expect-error
    row.querySelectorAll('td,th').forEach((/** @type {HTMLTableCellElement} */ col, c) => {
      const x = r - 1
      const z = c - +headerColumn
      const value = col.innerText

      if (x >= 0 && z >= 0) {
        zxValues[z] = zxValues[z] ?? []
        zxValues[z][x] = parseFloat(value)
      } else if (x >= 0 && z < 0) {
        xLabels[x] = value
      } else if (x < 0 && z >= 0) {
        zLabels[z] = value
      }
    }))

  if (reverse) {
    xLabels.reverse()
    zxValues.forEach(xValues => xValues.reverse())
  }

  const maxX = zxValues[0].length
  const xPx = Array.from({ length: maxX }, (_, x) => x * gapX)
  const yPx = Array.from({ length: maxY }, (_, y) => y * gapY)
  const width = gapX * (maxX - 1)
  const height = gapY * (maxY - 1)
  const svgW = left + width + right
  const svgH = top + height + bottom

  const /** @type {import('./h.js').HArgs1} */ gGraph = ['g', { transform: `translate(${left} ${top})` },
    ...yPx.map(px => ['line', { x1: 0, x2: width, y1: px, y2: px, stroke: '#8888' }]),
    ...xPx.map(px => ['line', { x1: px, x2: px, y1: 0, y2: height, stroke: '#8888' }])]
  const /** @type {import('./h.js').HArgs1} */ gZColors = ['g', { transform: `translate(${PADDING} ${top})` }]
  const /** @type {import('./h.js').HArgs1} */ gZLabels = ['g', { transform: `translate(${2 * PADDING} ${top})` }]
  const /** @type {import('./h.js').HArgs} */ svg = ['svg',
    { viewBox: `0 0 ${svgW} ${svgH}`, width: `${svgW}px`, height: `${svgH}px`, class: 'chartable' },
    gGraph, gZColors, gZLabels]

  if (xLabels.length > 0) {
    svg.push(['g', { transform: `translate(${left} ${top + height + PADDING})` },
      ...xPx.slice(0).map((xp, x) => ['text',
        rotate
          ? { x: 0, y: -xp, 'text-anchor': 'start', 'alignment-baseline': 'middle', transform: 'rotate(90)' }
          : { x: xp, y: PADDING, 'text-anchor': 'middle', 'alignment-baseline': 'middle' },
        xLabels[x]])])
  }

  const /** @type {[SVGElement, SVGElement, SVGTextElement][]} */ toUpdate = []

  const onclick = (/** @type {number} */ z) => toUpdate.forEach(([gLeftYLabels, gRightYLabels, zLabel], z2) => {
    s(gLeftYLabels, { $style: { display: z2 === z ? 'block' : 'none' } })
    s(gRightYLabels, { $style: { display: z2 === z ? 'block' : 'none' } })
    s(zLabel, { $style: { textDecoration: z2 === z ? 'underline' : 'none', cursor: 'pointer' } })
  })

  zxValues.forEach((values, z) => {
    const fill = COLORS[z % COLORS.length]
    const minValue = Math.min(...values.filter(value => !isNaN(value)))
    const maxValue = Math.max(...values.filter(value => !isNaN(value)))

    const valuesPx = values.map(value => isNaN(value) ? value : height * (maxValue - value) / (maxValue - minValue))

    const gLeftYLabels = s('g', { transform: `translate(${left - PADDING} ${top})` },
      ...yPx.map((yp, y) => ['text',
        { x: 0, y: yp, 'text-anchor': 'end', 'alignment-baseline': 'middle' },
        (maxValue - (maxValue - minValue) / maxY * (y + 1)).toFixed(2)]))

    const gRightYLabels = s('g', { transform: `translate(${left + width + PADDING} ${top})` },
      ...yPx.map((yp, y) => ['text',
        { x: 0, y: yp, 'text-anchor': 'start', 'alignment-baseline': 'middle' },
        (maxValue - (maxValue - minValue) / maxY * (y + 1)).toFixed(2)]))

    svg.push(gLeftYLabels, gRightYLabels)

    const zLabel = s('text', {
      x: 0,
      y: gapY * (gZLabels.length - 2),
      'text-anchor': 'start',
      'alignment-baseline': 'middle',
      $onclick: () => onclick(z)
    },
    zLabels[z])

    gZLabels.push(zLabel)
    gZColors.push(['circle', { cx: 0, cy: gapY * (gZColors.length - 2), r: 5, fill }])

    toUpdate.push([gLeftYLabels, gRightYLabels, zLabel])

    xPx.forEach((px, x) => {
      if (!isNaN(valuesPx[x])) {
        gGraph.push(['g', ['circle', { cx: px, cy: valuesPx[x], r: 5, fill }]])

        if (!isNaN(valuesPx[x - 1])) {
          gGraph.push(['line', { x1: px, x2: xPx[x - 1], y1: valuesPx[x], y2: valuesPx[x - 1], stroke: fill }])
        }
      }
    })
  })

  onclick(0)

  return s(...svg)
}

export const tests = {}
