/**
 * A helper that checks if the given argument is of a certain type.
 *
 * @template {abstract new (...args: any[]) => any} T
 *
 * @type {{
 *   (type: BigIntConstructor,  arg: any): arg is bigint;
 *   (type: BooleanConstructor, arg: any): arg is boolean;
 *   (type: NumberConstructor,  arg: any): arg is number;
 *   (type: StringConstructor,  arg: any): arg is string;
 *   (type: SymbolConstructor,  arg: any): arg is symbol;
 *   (type: undefined,          arg: any): arg is undefined | null;
 *   <T extends abstract new (...args: any[]) => any>(type: T, arg: any): arg is InstanceType<T>;
 * }}
 *
 * @return {arg is bigint | boolean | number | string | symbol | undefined | null | InstanceType<T>}
 */
export const is = (/** @type {T} */ type, /** @type {any} */ arg) => arg?.constructor === type

export const tests = {
  is: (/** @type {import('bun:test').Expect} */ expect) => {
    expect(is(Number, 42)).toBe(true)
    expect(is(Number, Number(42))).toBe(true)
    // eslint-disable-next-line no-new-wrappers
    expect(is(Number, new Number(42))).toBe(true)
    expect(is(Number, NaN)).toBe(true)
    expect(is(String, '42')).toBe(true)
    expect(is(String, String('42'))).toBe(true)
    // eslint-disable-next-line no-new-wrappers
    expect(is(String, new String('42'))).toBe(true)
    expect(is(Symbol, Symbol('42'))).toBe(true)
    expect(is(Symbol, Object(Symbol('42')))).toBe(true)
    expect(is(undefined, undefined)).toBe(true)
    expect(is(undefined, null)).toBe(true)
    expect(is(Object, {})).toBe(true)
    expect(is(Array, [])).toBe(true)
    expect(is(RegExp, /42/)).toBe(true)
    expect(is(Date, new Date(42))).toBe(true)
    expect(is(Set, new Set(['42', 42]))).toBe(true)
    expect(is(Map, new Map([[{ j: 42 }, { J: '42' }], [{ c: 42 }, { C: '42' }]]))).toBe(true)
  },

  'is: toString.call': (/** @type {import('bun:test').Expect} */ expect) => {
    const iz = (/** @type {any} */ type, /** @type {any} */ arg) => ({}).toString.call(arg).slice(8, -1) === type?.name

    class FooBar { }

    expect(is(FooBar, new FooBar())).toBe(true)
    expect(iz(FooBar, new FooBar())).toBe(false)

    expect(is(Object, new FooBar())).toBe(false)
    expect(iz(Object, new FooBar())).toBe(true)

    const fakeFooBar = {}

    // @ts-expect-error
    fakeFooBar[Symbol.toStringTag] = FooBar.name

    expect(is(FooBar, fakeFooBar)).toBe(false)
    expect(iz(FooBar, fakeFooBar)).toBe(true)

    expect(is(Object, fakeFooBar)).toBe(true)
    expect(iz(Object, fakeFooBar)).toBe(false)
  }
}
