# nnn

Jackens’ JavaScript helpers.

<sub>Version: <code class="version">2024.2.17</code></sub>

## Installation

```sh
bun i @jackens/nnn
```

or

```sh
npm i @jackens/nnn
```

## Usage

Bun (or Node.js):

```js
import { «something» } from '@jackens/nnn'
```

The browser (or Bun or Node.js):

```js
import { «something» } from './node_modules/@jackens/nnn/nnn.js'
```

## Exports

- `EscapeMap`: The type of arguments of the `escapeValues` and `escape` helpers.
- `HArgs`: The type of arguments of the `h` and `s` helpers.
- `HArgs1`: The type of arguments of the `h` and `s` helpers.
- `JcssNode`: The type of arguments of the `jcss` helper.
- `JcssRoot`: The type of arguments of the `jcss` helper.
- `chartable`: A helper for creating a chart based on a table (conf. <https://jackens.github.io/nnn/chartable/>).
- `eq`: A helper that checks equality of the given arguments.
- `escape`: A generic helper for escaping `values` by given `escapeMap` (in *TemplateStrings* flavor).
- `escapeValues`: A generic helper for escaping `values` by given `escapeMap`.
- `fixTypography`: A helper that implements typographic corrections specific to Polish typography.
- `h`: A lightweight [HyperScript](https://github.com/hyperhype/hyperscript)-style helper for creating and modifying `HTMLElement`s (see also `s`).
- `has`: A replacement for the `in` operator (not to be confused with the `for-in` loop) that works properly.
- `is`: A helper that checks if the given argument is of a certain type.
- `jcss`: A simple CSS-in-JS helper.
- `jsOnParse`: `JSON.parse` with “JavaScript turned on”.
- `locale`: Language translations helper.
- `nanolight`: A generic helper for syntax highlighting (see also `nanolightJs`).
- `nanolightJs`: A helper for highlighting JavaScript.
- `omit`: A helper that implements TypeScript’s `Omit` utility type.
- `pick`: A helper that implements TypeScript’s `Pick` utility type.
- `plUral`: A helper for choosing the correct singular and plural.
- `pro`: A helper that protects calls to nested properties by a `Proxy` that initializes non-existent values with an empty object.
- `refsInfo`: A helper that provides information about the given `refs`.
- `s`: A lightweight [HyperScript](https://github.com/hyperhype/hyperscript)-style helper for creating and modifying `SVGElement`s (see also `h`).
- `uuid1`: A helper that generates a UUID v1 identifier (with a creation timestamp).

### EscapeMap

```ts
export type EscapeMap = Map<any, (value?: any) => string>;
```

The type of arguments of the `escapeValues` and `escape` helpers.

### HArgs

```ts
export type HArgs = [string | Node, ...HArgs1[]];
```

The type of arguments of the `h` and `s` helpers.

### HArgs1

```ts
export type HArgs1 = Partial<Record<PropertyKey, any>> | null | undefined | Node | string | number | [string | Node, ...HArgs1[]];
```

The type of arguments of the `h` and `s` helpers.

### JcssNode

```ts
export type JcssNode = {
    [attributeOrSelector: string]: string | number | JcssNode | undefined;
};
```

The type of arguments of the `jcss` helper.

### JcssRoot

```ts
export type JcssRoot = Partial<Record<PropertyKey, JcssNode>>;
```

The type of arguments of the `jcss` helper.

### chartable

```ts
export function chartable({ bottom, gapX, gapY, headerColumn, left, maxY, reverse, right, rotate, table, top }: {
    bottom?: number;
    gapX?: number;
    gapY?: number;
    headerColumn?: boolean;
    left?: number;
    maxY?: number;
    reverse?: boolean;
    right?: number;
    rotate?: boolean;
    table: HTMLTableElement;
    top?: number;
}): Node;
```

A helper for creating a chart based on a table (conf. <https://jackens.github.io/nnn/chartable/>).

Options:
- `bottom`: bottom padding (for X axis labels)
- `gapX`: X axis spacing
- `gapY`: Y axis spacing
- `headerColumn`: flag indicating that `table` has a header column (with X axis labels)
- `left`: left padding
- `maxY`: number of Y axis lines
- `reverse`: flag to reverse all data series
- `right`: right padding (for data series labels)
- `rotate`: flag to rotate X axis labels
- `table`: `HTMLTableElement` to extract data, data series labels and X axis labels
- `top`: top padding

### eq

```ts
export function eq(x: any, y: any): boolean;
```

A helper that checks equality of the given arguments.

#### Usage Examples

```js
expect(eq(true, true)).toBe(true)
expect(eq(NaN, NaN)).toBe(true)
expect(eq(null, undefined)).toBe(false)
expect(eq(42, 42)).toBe(true)
expect(eq(42, new Number(42))).toBe(true)
expect(eq(42, Number(42))).toBe(true)
expect(eq(new Number(42), Number(42))).toBe(true)
expect(eq(42, '42')).toBe(false)
expect(eq('42', '42')).toBe(true)
expect(eq('42', new String('42'))).toBe(true)
expect(eq('42', String('42'))).toBe(true)
expect(eq(String('42'), new String('42'))).toBe(true)
expect(eq(/42/, /42/)).toBe(true)
expect(eq(/42/, /42/g)).toBe(false)
expect(eq(new Date(42), new Date(42))).toBe(true)
expect(eq(new Date(), new Date(42))).toBe(false)
expect(eq({ j: '42', c: 42 }, { c: 42, j: '42' })).toBe(true)
expect(eq([42, '42'], [42, '42'])).toBe(true)
expect(eq(new Set(['42', 42]), new Set([42, '42']))).toBe(true)
expect(eq(new Set(['42', 42]), new Set([42]))).toBe(false)
expect(eq(new Set([42, undefined]), new Set([42]))).toBe(false)
expect(eq(
  new Map([[{ j: 42 }, { J: '42' }], [{ c: 42 }, { C: '42' }]]),
  new Map([[{ c: 42 }, { C: '42' }], [{ j: 42 }, { J: '42' }]])
)).toBe(true)
expect(eq(
  new Map([[{ j: 42 }, { J: '42' }], [{ c: 42 }, { C: '42' }]]),
  new Map([[{ j: '42' }, { J: 42 }], [{ c: '42' }, { C: 42 }]])
)).toBe(false)
```

### escape

```ts
export function escape(escapeMap: EscapeMap, template: TemplateStringsArray, ...values: any[]): string;
```

A generic helper for escaping `values` by given `escapeMap` (in *TemplateStrings* flavor).

#### Usage Examples

```js
const /** @type {EscapeMap} */ escapeMap = new Map([
  [undefined, () => 'NULL'],
  [Array, (/** @type {any[]} */ values) => escapeValues(escapeMap, values).join(', ')],
  [Boolean, (/** @type {boolean} */ value) => `b'${+value}'`],
  [Date, (/** @type {Date} */ value) => `'${value.toISOString().replace(/^(.+)T(.+)\..*$/, '$1 $2')}'`],
  [Number, (/** @type {number} */ value) => `${value}`],
  [String, (/** @type {string} */ value) => `'${value.replace(/'/g, "''")}'`]
])

const sql = escape.bind(null, escapeMap)

const actual = sql`
  SELECT *
  FROM table_name
  WHERE column_name IN (${[true, null, undefined, 42, '42', "4'2", /42/, new Date(323325000000)]})`

const expected = `
  SELECT *
  FROM table_name
  WHERE column_name IN (b'1', NULL, NULL, 42, '42', '4''2', NULL, '1980-03-31 04:30:00')`

expect(actual).toEqual(expected)
```

### escapeValues

```ts
export function escapeValues(escapeMap: EscapeMap, values: any[]): string[];
```

A generic helper for escaping `values` by given `escapeMap`.

### fixTypography

```ts
export function fixTypography(node: Node): void;
```

A helper that implements typographic corrections specific to Polish typography.

#### Usage Examples

```js
const p = h('p', 'Pchnąć w tę łódź jeża lub ośm skrzyń fig (zob. https://pl.wikipedia.org/wiki/Pangram).')

fixTypography(p)

expect(p.innerHTML).toEqual(
  'Pchnąć <span style="white-space:nowrap">w </span>tę łódź jeża lub ośm skrzyń fig ' +
  '(zob. https://\u200Bpl.\u200Bwikipedia.\u200Borg/\u200Bwiki/\u200BPangram).')
```

### h

```ts
export function h<T extends keyof HTMLElementTagNameMap>(tag: T, ...args1: HArgs1[]): HTMLElementTagNameMap[T];
export function h<N extends Node>(node: N, ...args1: HArgs1[]): N;
export function h(tagOrNode: string | Node, ...args1: HArgs1[]): Node;
```

A lightweight [HyperScript](https://github.com/hyperhype/hyperscript)-style helper for creating and modifying `HTMLElement`s (see also `s`).

- The first argument of type `string` specifies the tag of the element to be created.
- The first argument of type `Node` specifies the element to be modified.
- All other arguments of type `Partial<Record<PropertyKey, any>>` are mappings of attributes and properties.
  Keys starting with `$` specify *properties* (without the leading `$`) to be set on the element being created or modified.
  (Note that `$` is not a valid attribute name character.)
  All other keys specify *attributes* to be set by `setAttribute`.
  An attribute equal to `false` causes the attribute to be removed by `removeAttribute`.
- All other arguments of type `null` or `undefined` are simply ignored.
- All other arguments of type `Node` are appended to the element being created or modified.
- All other arguments of type `string`/`number` are converted to `Text` nodes and appended to the element being created or modified.
- All other arguments of type `HArgs` are passed to `h` and the results are appended to the element being created or modified.

#### Usage Examples

```js
const b = h('b')

expect(b.outerHTML).toEqual('<b></b>')

const i = h('i', 'text')

h(b, i)

expect(i.outerHTML).toEqual('<i>text</i>')
expect(b.outerHTML).toEqual('<b><i>text</i></b>')

h(i, { $className: 'some class' })

expect(i.outerHTML).toEqual('<i class="some class">text</i>')
expect(b.outerHTML).toEqual('<b><i class="some class">text</i></b>')
```

```js
expect(h('span', 'text').outerHTML).toEqual('<span>text</span>')
expect(h('span', { $innerText: 'text' }).outerHTML).toEqual('<span>text</span>')
```

```js
expect(h('div', { style: 'margin:0;padding:0' }).outerHTML)
  .toEqual('<div style="margin:0;padding:0"></div>')
expect(h('div', { $style: 'margin:0;padding:0' }).outerHTML)
  .toEqual('<div style="margin: 0px; padding: 0px;"></div>')
expect(h('div', { $style: { margin: 0, padding: 0 } }).outerHTML)
  .toEqual('<div style="margin: 0px; padding: 0px;"></div>')
```

```js
const input1 = h('input', { value: 42 })
const input2 = h('input', { $value: '42' })

expect(input1.value).toEqual('42')
expect(input2.value).toEqual('42')

expect(input1.outerHTML).toEqual('<input value="42">')
expect(input2.outerHTML).toEqual('<input>')

const checkbox1 = h('input', { type: 'checkbox', checked: true })
const checkbox2 = h('input', { type: 'checkbox', $checked: true })

expect(checkbox1.checked).toBe(true)
expect(checkbox2.checked).toBe(true)

expect(checkbox1.outerHTML).toEqual('<input type="checkbox" checked="">')
expect(checkbox2.outerHTML).toEqual('<input type="checkbox">')
```

```js
const div = h('div')

expect(div.key).toBeUndefined()

h(div, { $key: { one: 1 } })

expect(div.key).toEqual({ one: 1 })

h(div, { $key: { two: 2 } })

expect(div.key).toEqual({ one: 1, two: 2 })
```

### has

```ts
export function has(key: any, ref: any): boolean;
```

A replacement for the `in` operator (not to be confused with the `for-in` loop) that works properly.

#### Usage Examples

```js
const obj = { key: 'K', null: 'N' }

expect('key' in obj).toBe(true)
expect(has('key', obj)).toBe(true)

expect('null' in obj).toBe(true)
expect(has('null', obj)).toBe(true)

expect(null in obj).toBe(true)
expect(has(null, obj)).toBe(false)

expect('toString' in obj).toBe(true)
expect(has('toString', obj)).toBe(false)
```

```js
let typeError

try {
  'key' in null
} catch (error) {
  typeError = error
}

expect(typeError instanceof TypeError) // Cannot use 'in' operator to search for 'key' in null
expect(has('key', null)).toBe(false)
```

### is

```ts
export const is: {
    (type: BigIntConstructor, arg: any): arg is bigint;
    (type: BooleanConstructor, arg: any): arg is boolean;
    (type: NumberConstructor, arg: any): arg is number;
    (type: StringConstructor, arg: any): arg is string;
    (type: SymbolConstructor, arg: any): arg is symbol;
    (type: undefined, arg: any): arg is null | undefined;
    <T extends abstract new (...args: any[]) => any>(type: T, arg: any): arg is InstanceType<T>;
};
```

A helper that checks if the given argument is of a certain type.

#### Usage Examples

```js
expect(is(Number, 42)).toBe(true)
expect(is(Number, Number(42))).toBe(true)
expect(is(Number, new Number(42))).toBe(true)
expect(is(Number, NaN)).toBe(true)
expect(is(String, '42')).toBe(true)
expect(is(String, String('42'))).toBe(true)
expect(is(String, new String('42'))).toBe(true)
expect(is(Symbol, Symbol('42'))).toBe(true)
expect(is(Symbol, Object(Symbol('42')))).toBe(true)
expect(is(undefined, undefined)).toBe(true)
expect(is(undefined, null)).toBe(true)
expect(is(Object, {})).toBe(true)
expect(is(Array, [])).toBe(true)
expect(is(RegExp, /42/)).toBe(true)
expect(is(Date, new Date(42))).toBe(true)
expect(is(Set, new Set(['42', 42]))).toBe(true)
expect(is(Map, new Map([[{ j: 42 }, { J: '42' }], [{ c: 42 }, { C: '42' }]]))).toBe(true)
```

```js
const iz = (/** @type {any} */ type, /** @type {any} */ arg) => ({}).toString.call(arg).slice(8, -1) === type?.name

class FooBar { }

expect(is(FooBar, new FooBar())).toBe(true)
expect(iz(FooBar, new FooBar())).toBe(false)

expect(is(Object, new FooBar())).toBe(false)
expect(iz(Object, new FooBar())).toBe(true)

const fakeFooBar = {}

fakeFooBar[Symbol.toStringTag] = FooBar.name

expect(is(FooBar, fakeFooBar)).toBe(false)
expect(iz(FooBar, fakeFooBar)).toBe(true)

expect(is(Object, fakeFooBar)).toBe(true)
expect(iz(Object, fakeFooBar)).toBe(false)
```

### jcss

```ts
export function jcss(root: JcssRoot, splitter?: string): string;
```

A simple CSS-in-JS helper.

The `root` parameter provides a hierarchical description of CSS rules.

- Keys of sub-objects whose values are NOT objects are treated as CSS attribute, and values are treated as values of those CSS attributes; the concatenation of keys of all parent objects is a CSS rule.
- All keys ignore the part starting with a splitter (default: `$$`) sign until the end of the key (e.g. `src$$1` → `src`, `@font-face$$1` → `@font-face`).
- In keys specifying CSS attribute, all uppercase letters are replaced by lowercase letters with an additional `-` character preceding them (e.g. `fontFamily` → `font-family`).
- Commas in keys that makes a CSS rule cause it to “split” and create separate rules for each part (e.g. `{div:{margin:1,'.a,.b,.c':{margin:2}}}` → `div{margin:1}div.a,div.b,div.c{margin:2}`).
- Top-level keys that begin with `@` are not concatenated with sub-object keys.

#### Usage Examples

```js
const actual = jcss({
  a: {
    color: 'red',
    margin: 1,
    '.c': { margin: 2, padding: 2 },
    padding: 1
  }
})

const expected = `
a{
  color:red;
  margin:1
}
a.c{
  margin:2;
  padding:2
}
a{
  padding:1
}`.replace(/\n\s*/g, '')

expect(actual).toEqual(expected)
```

```js
const actual = jcss({
  a: {
    '.b': {
      color: 'red',
      margin: 1,
      '.c': { margin: 2, padding: 2 },
      padding: 1
    }
  }
})

const expected = `
a.b{
  color:red;
  margin:1
}
a.b.c{
  margin:2;
  padding:2
}
a.b{
  padding:1
}`.replace(/\n\s*/g, '')

expect(actual).toEqual(expected)
```

```js
const actual = jcss({
  '@font-face$$1': {
    fontFamily: 'Jackens',
    src$$1: 'url(otf/jackens.otf)',
    src$$2: "url(otf/jackens.otf) format('opentype')," +
      "url(svg/jackens.svg) format('svg')",
    fontWeight: 'normal',
    fontStyle: 'normal'
  },
  '@font-face$$2': {
    fontFamily: 'C64',
    src: 'url(fonts/C64_Pro_Mono-STYLE.woff)'
  },
  '@keyframes spin': {
    '0%': { transform: 'rotate(0deg)' },
    '100%': { transform: 'rotate(360deg)' }
  },
  div: {
    border: 'solid red 1px',
    '.c1': { 'background-color': '#000' },
    ' .c1': { backgroundColor: 'black' },
    '.c2': { backgroundColor: 'rgb(0,0,0)' }
  },
  '@media(min-width:200px)': {
    div: { margin: 0, padding: 0 },
    span: { color: '#000' }
  }
})

const expected = `
@font-face{
  font-family:Jackens;
  src:url(otf/jackens.otf);
  src:url(otf/jackens.otf) format('opentype'),url(svg/jackens.svg) format('svg');
  font-weight:normal;
  font-style:normal
}
@font-face{
  font-family:C64;
  src:url(fonts/C64_Pro_Mono-STYLE.woff)
}
@keyframes spin{
  0%{
    transform:rotate(0deg)
  }
  100%{
    transform:rotate(360deg)
  }
}
div{
  border:solid red 1px
}
div.c1{
  background-color:#000
}
div .c1{
  background-color:black
}
div.c2{
  background-color:rgb(0,0,0)
}
@media(min-width:200px){
  div{
    margin:0;
    padding:0
  }
  span{
    color:#000
  }
}`.replace(/\n\s*/g, '')

expect(actual).toEqual(expected)
```

```js
const actual = jcss({
  a: {
    '.b,.c': {
      margin: 1,
      '.d': {
        margin: 2
      }
    }
  }
})

const expected = `
a.b,a.c{
  margin:1
}
a.b.d,a.c.d{
  margin:2
}`.replace(/\n\s*/g, '')

expect(actual).toEqual(expected)
```

```js
const actual = jcss({
  '.b,.c': {
    margin: 1,
    '.d': {
      margin: 2
    }
  }
})

const expected = `
.b,.c{
  margin:1
}
.b.d,.c.d{
  margin:2
}`.replace(/\n\s*/g, '')

expect(actual).toEqual(expected)
```

```js
const actual = jcss({
  '.a,.b': {
    margin: 1,
    '.c,.d': {
      margin: 2
    }
  }
})

const expected = `
.a,.b{
  margin:1
}
.a.c,.a.d,.b.c,.b.d{
  margin:2
}`.replace(/\n\s*/g, '')

expect(actual).toEqual(expected)
```

### jsOnParse

```ts
export function jsOnParse(handlers: Partial<Record<PropertyKey, Function>>, text: string): any;
```

`JSON.parse` with “JavaScript turned on”.

Objects having *exactly* one property which is present in the `handlers` map, i.e. objects of the form:

```js
{ "«handlerName»": [«params»] }
```

are replaced by the result of call

```js
handlers['«handlerName»'](...«params»)
```

#### Usage Examples

```js
const handlers = {
  $hello: (/** @type {string} */ name) => `Hello ${name}!`,
  $foo: () => 'bar'
}
const actual = jsOnParse(handlers, `[
  {
    "$hello": ["World"]
  },
  {
    "nested": {
      "$hello": ["nested World"]
    },
    "one": 1,
    "two": 2
  },
  {
    "$foo": []
  },
  {
    "$foo": ["The parent object does not have exactly one property!"],
    "one": 1,
    "two": 2
  }
]`)
const expected = [
  'Hello World!',
  {
    nested: 'Hello nested World!',
    one: 1,
    two: 2
  },
  'bar',
  {
    $foo: ['The parent object does not have exactly one property!'],
    one: 1,
    two: 2
  }
]

expect(actual).toEqual(expected)
```

### locale

```ts
export function locale(map: Partial<Record<PropertyKey, Partial<Record<PropertyKey, string>>>>, defaultVersion: string): (text: string, version?: string) => string;
```

Language translations helper.

#### Usage Examples

```js
const _ = locale({
  default: { Password: 'Hasło' },
  button: { Login: 'Zaloguj' }
}, 'default')

expect(_('Login')).toEqual('Login')
expect(_('Password')).toEqual('Hasło')

expect(_('Undefined text')).toEqual('Undefined text')

expect(_('Login', 'button')).toEqual('Zaloguj')

expect(_('Password', 'undefined_version')).toEqual('Hasło')
expect(_('Undefined text', 'undefined_version')).toEqual('Undefined text')

expect(_('toString')).toEqual('toString')
expect(_('toString', 'undefined_version')).toEqual('toString')
```

### nanolight

```ts
export function nanolight(pattern: RegExp, highlighters: ((chunk: string, index: number) => import('./h.js').HArgs1)[], code: string): import("./h.js").HArgs1[];
```

A generic helper for syntax highlighting (see also `nanolightJs`).

### nanolightJs

```ts
export const nanolightJs: (code?: string | undefined) => import("./h.js").HArgs1[];
```

A helper for highlighting JavaScript.

#### Usage Examples

```js
const codeJs = 'const answerToLifeTheUniverseAndEverything = 42'

expect(h('pre', ['code', ...nanolightJs(codeJs)]).outerHTML).toEqual(
  '<pre><code><span class="keyword">const</span> <span class="literal">answerToLifeTheUniverseAndEverything</span> <span class="operator">=</span> <span class="number">42</span></code></pre>')
```

### omit

```ts
export const omit: <T extends Partial<Record<PropertyKey, any>>, K extends (keyof T)[]>(obj: T, keys: K) => Omit<T, K[number]>;
```

A helper that implements TypeScript’s `Omit` utility type.

#### Usage Examples

```js
const obj = { a: 42, b: '42', c: 17 }

expect(omit(obj, ['c'])).toEqual({ a: 42, b: '42' })
```

### pick

```ts
export const pick: <T extends Partial<Record<PropertyKey, any>>, K extends (keyof T)[]>(obj: T, keys: K) => Pick<T, K[number]>;
```

A helper that implements TypeScript’s `Pick` utility type.

#### Usage Examples

```js
const obj = { a: 42, b: '42', c: 17 }

expect(pick(obj, ['a', 'b'])).toEqual({ a: 42, b: '42' })
```

### plUral

```ts
export function plUral(singular: string, plural2: string, plural5: string, value: number): string;
```

A helper for choosing the correct singular and plural.

#### Usage Examples

```js
const auto = plUral.bind(null, 'auto', 'auta', 'aut')

expect(auto(0)).toEqual('aut')
expect(auto(1)).toEqual('auto')
expect(auto(17)).toEqual('aut')
expect(auto(42)).toEqual('auta')

const car = plUral.bind(null, 'car', 'cars', 'cars')

expect(car(0)).toEqual('cars')
expect(car(1)).toEqual('car')
expect(car(17)).toEqual('cars')
expect(car(42)).toEqual('cars')
```

### pro

```ts
export function pro(ref: any): any;
```

A helper that protects calls to nested properties by a `Proxy` that initializes non-existent values with an empty object.

#### Usage Examples

```js
const ref = {}

pro(ref).one.two[3][4] = 1234

expect(ref).toEqual({ one: { two: { 3: { 4: 1234 } } } })

pro(ref).one.two.tree = 123

expect(ref).toEqual({ one: { two: { 3: { 4: 1234 }, tree: 123 } } })

pro(ref).one.two = undefined

expect(ref).toEqual({ one: { two: undefined } })

delete pro(ref).one.two

expect(ref).toEqual({ one: {} })

pro(ref).one.two.three.four

expect(ref).toEqual({ one: { two: { three: { four: {} } } } })

pro(ref).one.two.three.four = 1234

expect(ref).toEqual({ one: { two: { three: { four: 1234 } } } })
```

### refsInfo

```ts
export function refsInfo(...refs: any[]): [string, string, string[]][];
```

A helper that provides information about the given `refs`.

It returns an array of triples: `[«name», «prototype-name», «array-of-own-property-names»]`.

#### Usage Examples

```js
const info = refsInfo(Array, Function)

expect(info.find(([name]) => name === 'Array')?.[2]?.includes('length')).toBe(true)
expect(info.find(([name]) => name === 'Function')?.[2]?.includes('length')).toBe(true)
```

```js
const browserFingerprint = () => {
  const refs = Object.getOwnPropertyNames(window).map(name => window[name])
  const info = refsInfo(...refs)
  const json = JSON.stringify(info)
  const hash = Array(32).fill(0)
  let j = 0

  for (let i = 0; i < json.length; i++) {
    let charCode = json.charCodeAt(i)

    while (charCode > 0) {
      hash[j] = hash[j] ^ (charCode & 15)
      charCode >>= 4
      j = (j + 1) & 31
    }
  }

  return hash.map(x => x.toString(16)).join('')
}

console.log(browserFingerprint())
```

### s

```ts
export function s<T extends keyof SVGElementTagNameMap>(tag: T, ...args1: HArgs1[]): SVGElementTagNameMap[T];
export function s<N extends Node>(node: N, ...args1: HArgs1[]): N;
export function s(tagOrNode: string | Node, ...args1: HArgs1[]): Node;
```

A lightweight [HyperScript](https://github.com/hyperhype/hyperscript)-style helper for creating and modifying `SVGElement`s (see also `h`).

- The first argument of type `string` specifies the tag of the element to be created.
- The first argument of type `Node` specifies the element to be modified.
- All other arguments of type `Partial<Record<PropertyKey, any>>` are mappings of attributes and properties.
  Keys starting with `$` specify *properties* (without the leading `$`) to be set on the element being created or modified.
  (Note that `$` is not a valid attribute name character.)
  All other keys specify *attributes* to be set by `setAttributeNS`.
  An attribute equal to `false` causes the attribute to be removed by `removeAttributeNS`.
- All other arguments of type `null` or `undefined` are simply ignored.
- All other arguments of type `Node` are appended to the element being created or modified.
- All other arguments of type `string`/`number` are converted to `Text` nodes and appended to the element being created or modified.
- All other arguments of type `HArgs` are passed to `s` and the results are appended to the element being created or modified.

### uuid1

```ts
export function uuid1({ date, node }?: {
    date?: Date | undefined;
    node?: string | undefined;
}): string;
```

A helper that generates a UUID v1 identifier (with a creation timestamp).

- The optional `node` parameter should have the format `/^[0123456789abcdef]+$/`.
  Its value will be trimmed to last 12 characters and left padded with zeros.

#### Usage Examples

```js
for (let i = 1; i <= 22136; ++i) {
  const uuid = uuid1()

  i === 1 && expect(uuid.split('-')[3]).toEqual('8001')
  i === 4095 && expect(uuid.split('-')[3]).toEqual('8fff')
  i === 4096 && expect(uuid.split('-')[3]).toEqual('9000')
  i === 9029 && expect(uuid.split('-')[3]).toEqual('a345')
  i === 13398 && expect(uuid.split('-')[3]).toEqual('b456')
  i === 16384 && expect(uuid.split('-')[3]).toEqual('8000')
  i === 17767 && expect(uuid.split('-')[3]).toEqual('8567')
}
```

```js
expect(uuid1({ node: '000123456789abc' }).split('-')[4]).toEqual('123456789abc')
expect(uuid1({ node: '123456789' }).split('-')[4]).toEqual('000123456789')
```

```js
expect(uuid1({ date: new Date(323325000000) })).toStartWith('c1399400-9a71-11bd')
```

## License

The MIT License (MIT)

Copyright (c) 2016+ Jackens

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
